/*
 * GameMap.java
 *
 * Created on August 30, 2002, 11:41 PM
 */
import java.util.*;

/**
 *
 * @author  scerruti
 */
public class GameMap {
    
    private int width;
    private int height;
    private GameTile[][] gameBoard;
    private char[][] botMap = null;
    private HashSet bases;
    private HashSet permBases;
    private HashMap robots;
    
    public static final char NORMAL = '.';
    public static final char WALL   = '#';
    public static final char BASE   = '@';
    public static final char WATER  = '~';
    public static final char ROBOT  = 'R';
    
    private class GameTile {
        
        private char type;
        
        public GameTile(char type) throws BadTileTypeException {
            super();
            switch (type) {
                case NORMAL:
                case WALL:
                case BASE:
                case WATER:
                case ROBOT:
                    this.type=type;
                    break;
                default:
                    throw new BadTileTypeException();
            }
        }
        
        char getType() {
            return (type);
        }
        
        void setType (char type) {
            this.type = type;
        }
               
        void dump() {
            System.out.print(this.type);
        }
    }
    
    /** Creates a new instance of GameMap */
    public GameMap(int width, int height) {
        super();
        this.width  = width;
        this.height = height;
        this.gameBoard   = new GameTile[width+1][height+1];
        this.bases = new HashSet();
        this.permBases = new HashSet();
        this.robots = new HashMap ();
    }
    
    void addLine(int row, String line) throws BadTileTypeException {
        //System.out.println(row+": map line> "+line);
        for (int i=1; i <= width; i++) {
            gameBoard[i][row] = new GameTile(line.charAt(i-1));
            if (gameBoard[i][row].getType() == BASE) {
                GameBoardLocation loc = new GameBoardLocation (this, i, row);
                bases.add(loc);
                permBases.add(loc);
            }
        }
    }
    
    int getHeight () {
        return (this.height);
    }
    
    int getWidth () {
        return (this.width);
    }
    
    void addBase(int x, int y) {
        if (gameBoard[x][y].getType() != BASE) {
            bases.add(new GameBoardLocation(this,x,y));
            gameBoard[x][y].setType(BASE);
            //System.out.println("ab: "+x+","+y);
        }
    }
    
    void addBase(GameBoardLocation loc) {
        addBase(loc.getX(), loc.getY());
    }
    
    void reloadBases() {
        bases.addAll(permBases);
    }
    
    void clearBase(int x, int y) {
        if (gameBoard[x][y].getType() == BASE) {
            // Continuous Server Mode
            //if (bases.size() == 1) {
                // If we are deleting the last base, readd all the original bases
                //bases.clear();
                //reloadBases();
            //}
            Iterator i = bases.iterator();
            while (i.hasNext()) {
                GameBoardLocation loc = (GameBoardLocation) i.next();
                if (loc.getX() == x && loc.getY() ==y) {
                    i.remove();
                    break;
                }
            }
            gameBoard[x][y].setType(NORMAL);
            //System.out.println("cb: "+x+","+y);
        }
    }

    void clearBase(GameBoardLocation loc) {
        clearBase(loc.getX(), loc.getY());
    }
    
    void doRobotCommand(RobotServerCommand rsc) {
        switch (rsc.command) {
            case RobotServerCommand.MOVE_EAST:
                moveRobot(rsc.id, MapSolver.EAST);
                break;
            case RobotServerCommand.MOVE_WEST:
                moveRobot(rsc.id, MapSolver.WEST);
                break;
            case RobotServerCommand.MOVE_NORTH:
                moveRobot(rsc.id, MapSolver.NORTH);
                break;
            case RobotServerCommand.MOVE_SOUTH:
                moveRobot(rsc.id, MapSolver.SOUTH);
                break;
            case RobotServerCommand.SET_LOCATION:
                try {
                    addRobot(new PkgRobot(this, rsc.id, rsc.x, rsc.y));
                } catch (BadLocationException e) {
                }
                break;
            case RobotServerCommand.NO_COMMAND:
            default:
                PkgRobot bot = (PkgRobot) robots.get(new Integer(rsc.id));
                bot.setSeen();
                break;
        }
    }
    
    void addRobot(PkgRobot robot) {
        robots.put(new Integer(robot.getId()), robot);
    }
    
    void moveRobot(int id, char direction) {
        PkgRobot bot = (PkgRobot) robots.get(new Integer(id));
        try {
            switch (direction) {
                case MapSolver.EAST:
                    bot.getLocation().moveEast();
                    break;
                case MapSolver.WEST:
                    bot.getLocation().moveWest();
                    break;
                case MapSolver.NORTH:
                    bot.getLocation().moveNorth();
                    break;
                case MapSolver.SOUTH:
                    bot.getLocation().moveSouth();
                    break;
            }
            bot.setSeen();
        } catch (BadLocationException e) {
            robots.remove(new Integer(id));
        }
    }
    
    void cleanRobots() {
        if (robots == null) return;
               
        botMap = new char[width+1][height+1];
        for (int i = 1; i <= width; i++) {
            for (int j = 1; j <= height; j++) {
                botMap[i][j] = NORMAL;
            }
        }
        
        Iterator i = robots.entrySet().iterator();
        while (i.hasNext()) {
            Map.Entry e = (Map.Entry) i.next();
            PkgRobot bot = (PkgRobot) e.getValue();
            if (bot.wasSeen()) {
                botMap[bot.getLocation().getX()][bot.getLocation().getY()] =
                    ROBOT;
                bot.clearSeen();
                continue;
            } else {
                i.remove();
            }
        }
    }
      
    char tileTypeAt(int x, int y) {
        if (x < 1 || x > getWidth() || y < 1 || y > getWidth()) return WATER;
        return (gameBoard[x][y].getType());
    }
    
    int dangerAt(int x, int y) {
        int danger = 0;
        
        if (x < 1 || x > getWidth() || y < 1 || y > getHeight()) {
            danger = danger+5;
        } else {
            // Environmental Danger
            switch (tileTypeAt(x,y)) {
                case WALL:
                    danger = danger+2;
                    break;
                case WATER:
                    danger = danger+5;
                    break;
            }
            
            if (botMap != null) {
                switch (botMap[x][y]) {
                    case ROBOT:
                        danger = danger + 10;
                        break;
                }
            }
        }
        
        // Robot Danger
        return (danger);
    }
    
    HashSet getBases() {
        return this.bases;
    }
    
    HashMap getRobots() {
        return this.robots;
    }
    
    void dump () {
        System.out.println(width+" "+height);
        for (int i = 1; i <= height; i++) {
            for (int j = 1; j <= width; j++) {
                gameBoard[j][i].dump();
            }
            if (botMap != null) {
                System.out.print("\t");
                for (int j = 1; j <= width; j++) {
                    System.out.print(botMap[j][i]);
                }
            }
            System.out.println();
        }
        if (robots != null) {
            Iterator iter = robots.entrySet().iterator();
            while (iter.hasNext()) {
                Map.Entry e = (Map.Entry) iter.next();
                System.out.print(e.getKey()+"> ");
                ((PkgRobot) e.getValue()).getLocation().dump();
            }
        }
    }
}
